@@ -633,7 +633,11 @@ osmo_sccp_simple_client_on_ss7_id(void *ctx, uint32_t ss7_id, const char *name,
 
 	/* Ensure that the ASP we use is set to client mode. */
 	asp->cfg.is_server = false;
-	asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
+
+	/* Make sure that the role of this ASP is set to ASP unless the user
+	 * made a concious decision about the role via the VTY */
+	if (!asp->cfg.role_set_by_vty)
+		asp->cfg.role = OSMO_SS7_ASP_ROLE_ASP;
 
 	/* Restart ASP */
 	if (prot != OSMO_SS7_ASP_PROT_IPA)
